# CREATE INDEX index_name ON table.name (name)
# [{'CreateIndex': {'name': [{'value': 'index', 'quote_style': None}], 'table_name': [{'value': 'table', 'quote_style': None}, {'value': 'name', 'quote_style': None}], 'columns': [{'expr': {'Identifier': {'value': 'name', 'quote_style': None}}, 'asc': None, 'nulls_first': None}], 'unique': False, 'if_not_exists': False}}]



class QueryStatistics:
    def __init__(self):
        self.count_blobs_found: int = 0
        self.count_data_blobs_read: int = 0
        self.bytes_read_data: int = 0
        self.bytes_processed_data: int = 0
        self.rows_read: int = 0

        self.partitions_found: int = 0
        self.partitions_scanned: int = 0
        self.partitions_read: int = 0

        self.time_data_read: int = 0

        # time spent query planning
        self.time_planning: int = 0

        self.start_time: int = 0
        self.end_time: int = 0

    def merge(self, stats):
        pass

    def as_dict(self):
        return {
            "count_blobs_found": self.count_blobs_found,
            "count_data_blobs_read": self.count_data_blobs_read,
            "bytes_read_data": self.bytes_read_data,
            "bytes_processed_data": self.bytes_processed_data,
            "rows_read": self.rows_read,
            "time_data_read": 0
            if self.time_data_read == 0
            else (self.time_data_read / 1e9),
            "time_total": (self.end_time - self.start_time) / 1e9,
            "time_planning": self.time_planning / 1e9,
            "partitions_found": self.partitions_found,
            "partitions_scanned": self.partitions_scanned,
            "partitions_read": self.partitions_read,
        }



"""
Evaluation Node

This is a SQL Query Execution Plan Node.

"""
from typing import Iterable
from opteryx.engine.query_statistics import QueryStatistics
from opteryx.engine.planner.operations.base_plan_node import BasePlanNode


class EvaluationNode(BasePlanNode):
    def __init__(self, statistics: QueryStatistics, **config):
        self._distinct = config.get("distinct", True)

    def execute(self, data_pages: Iterable) -> Iterable:

        # for each of the items that require a calculation, do it

        yield from data_pages



"""
Grouping Node

This is a SQL Query Execution Plan Node.

This performs aggregations.

This is a greedy operator - it consumes all the data before responding.
"""
import io
import orjson
import pyarrow.json
import numpy as np
from typing import Iterable
from opteryx.third_party.pyarrow_ops.group import Grouping
from opteryx.engine import QueryStatistics
from opteryx.engine.planner.operations import BasePlanNode

# these functions can be applied to each group
INCREMENTAL_AGGREGATES = {
    "MIN": lambda x, y: min(np.min(x), y),
    "MAX": lambda x, y: max(np.max(x), y),
    "SUM": lambda x, y: np.sum(x) + y,
    "COUNT": lambda x, y: np.size(x) + y,
}

# incremental updates usually need a suitable start
INCREMENTAL_AGGREGATES_SEEDS = {
    "MIN": np.min,
    "MAX": np.max,
    "SUM": np.sum,
    "COUNT": np.size,
}

# these functions need the whole dataset
WHOLE_AGGREGATES = {
    "AVG": np.mean,
    "MEDIAN": np.median,
    "PRODUCT": np.prod,
    "STDDEV_POP": np.std,
    "VAR_POP": np.var,
    "FIRST": lambda a: a[0],
    "LAST": lambda a: a[-1],
    # range - difference between min and max
    # percent - each group has the relative portion calculated
    # list - return a list of the items in the list
    # quantile
    # approx_distinct
    # approx_quantile
}


def _incremental(x, y, function):
    if function in INCREMENTAL_AGGREGATES:
        return INCREMENTAL_AGGREGATES[function](x, y)
    return np.concatenate(x, y)


def _incremental_seed(x, function):
    if function in INCREMENTAL_AGGREGATES_SEEDS:
        return INCREMENTAL_AGGREGATES_SEEDS[function](x)
    return x


def _serializer(obj):
    if isinstance(obj, np.int64):
        return int(obj)


class AggregateNode(BasePlanNode):
    def __init__(self, statistics: QueryStatistics, **config):
        self._aggregates = []
        self._project = config.get("groups", [])
        aggregates = config.get("aggregates", [])
        print(aggregates)
        for attribute in aggregates:
            if "aggregate" in attribute:
                self._aggregates.append(attribute)
            else:
                self._project.append(attribute)

        self._project = list(set(self._project))

    def __repr__(self):
        return str(self._aggregates)

    def greedy(self):
        return True

    def execute(self, groups: Iterable) -> Iterable:

        from collections import defaultdict

        collector = defaultdict(dict)

        # if we're not running this after a Group By, we need to add a layer
        if not isinstance(groups, Grouping):
            groups = [groups]

        for page in groups:
            for group in page:
                # We build the key value for the group collector
                group_identifier = tuple(
                    [(col, group[col][0]) for col in self._project]
                )

                for aggregrator in self._aggregates:

                    attribute = aggregrator["args"][0][0]
                    function = aggregrator["aggregate"]
                    column_name = f"{function}({attribute})"

                    group_collector = collector[group_identifier]

                    # Add the responses to the collector
                    # if it's COUNT(*) - we have a shortcut
                    if column_name == "COUNT(Wildcard)":
                        count = group_collector.get("COUNT(*)", 0)
                        group_collector["COUNT(*)"] = count + group.num_rows
                    # if we have some information collected for this group,
                    # incremental update
                    elif column_name in group_collector:
                        group_collector[column_name] = _incremental(
                            group[attribute],
                            group_collector[column_name],
                            function,
                        )
                    # otherwise, it's new, so seed the collection
                    else:
                        group_collector[column_name] = _incremental_seed(
                            group[attribute], function
                        )

                    collector[group_identifier] = group_collector
                # TODO: if we're going to cap the number of groups we collect, do it here

        # TODO: do any whole aggregate functions

        import time

        buffer = bytearray()
        t = time.time_ns()
        for collected, record in collector.items():
            # we can't load huge json docs into pyarrow, so we chunk it
            if len(buffer) > (1024 * 1024) :  # 1Mb - the default page size in mabel
                table = pyarrow.json.read_json(io.BytesIO(buffer))
                yield table
                buffer = bytearray()
            for k, v in collected:
                if hasattr(v, "as_py"):
                    v = v.as_py()
                record[k] = v
            buffer.extend(orjson.dumps(record, default=_serializer))

        if len(buffer) > 0:
            table = pyarrow.json.read_json(io.BytesIO(buffer))
            yield table

        print("building group table", (time.time_ns() - t) / 1e9)



"""
Grouping Node

This is a SQL Query Execution Plan Node.

This performs the grouping, but not the aggregations.
"""
from typing import Iterable
from opteryx.engine import QueryStatistics
from opteryx.engine.planner.operations import BasePlanNode
from opteryx.utils.pyarrow import groupby


class GroupNode(BasePlanNode):
    def __init__(self, statistics: QueryStatistics, **config):
        self._groups = config.get("groups", [])

    def __repr__(self):
        return str(self._groups)

    def execute(self, data_pages: Iterable) -> Iterable:

        for page in data_pages:
            yield from groupby(page, self._groups)



# cython: language_level=3
"""
This module is compiled, any changes to it need the following to be run before they
will be effective:

python setup.py build_ext --inplace
"""
import cython
from cityhash import CityHash32
from collections import defaultdict
from opteryx.engine.indexes.accumulation_tree import AccumulationTree


AGGREGATORS = {
    "SUM": lambda x, y: x + y,
    "MAX": max,
    "MIN": min,
    "COUNT": lambda x, y: x + 1,
    "AVG": lambda x, y: 1,
}


class TooManyGroups(Exception):
    pass


class GroupBy:
    """
    GroupBy does a lazy evaluation of the groups, the groups are calculated as part of
    calculating the aggregations. This was implemented like this so that generators
    can be aggregated - we have one opportunity to cycle of the records, and if the
    data is in a generator, there's a chance the dataset doesn't fit in memory.
    """

    def __init__(self, dictset, columns):
        self._dictset = dictset
        if isinstance(columns, (list, set, tuple)):
            self._columns = tuple(columns)
        else:
            self._columns = [columns]
        self._group_keys = {}

    def _map(self, collect_columns):
        """
        Create Tuples of records in the Groups (GroupID, CollectedColumn, Value)

        The GroupID is a hash of the grouped columns, we do this because we don't actually
        care about the column values, just that we can uniquely identify records with
        the same values.

        For each column we're collecting, we emit a record of the column and the value
        in the column.

        This is akin to the MAP step in a MapReduce algo, we're creating a set of values
        which standardize the format of the data to be processed and could allow the
        data to be processed in parallel.
        """
        if collect_columns == self._columns == {"*"}:
            # if we're doing COUNT(*), short-cut the processing
            self._group_keys["*"] = [("*", "*")]
            for record in self._dictset:
                yield ("*", "*", "*")
            return

        for record in self._dictset:
            try:
                group_key: cython.ulong = CityHash32(
                    "".join([str(record[column]) for column in self._columns]),
                )
            except KeyError:
                group_key: cython.ulong = CityHash32(
                    "".join([f"{record.get(column, '')}" for column in self._columns]),
                )
            if group_key not in self._group_keys.keys():
                self._group_keys[group_key] = [
                    (column, record.get(column)) for column in self._columns
                ]
                if len(self._group_keys) >= 4999999:
                    raise TooManyGroups(
                        f"Groups are not selective enough and too many Groups have been found (stopped at {len(self._group_keys)})."
                    )

            for column in collect_columns:
                if column == "*":
                    yield (group_key, column, "*")
                else:
                    v = record.get(column)  # ignore nulls
                    if v is not None:
                        yield (group_key, column, record[column])

    def aggregate(self, aggregations):
        """
        This implements steps akin to the REDUCE step in MapReduce.

        We work out with group to to map the result to and then REDUCE the resulant
        value from the set.

        This isn't intended for internal use only, but if you know how, you can
        call it.
        """

        if not isinstance(aggregations, list):
            aggregations = [aggregations]
        if not all(isinstance(agg, tuple) for agg in aggregations):
            raise ValueError("`aggregate` expects a list of Tuples")

        requested_aggs = aggregations.copy()

        # averages need the sum and the count
        for func, col in aggregations:
            if func == "AVG":
                aggregations += [("SUM", col), ("COUNT", col)]

        columns_to_collect = {col for func, col in aggregations}

        collector = AccumulationTree()
        # Iterate through the data in the groups formatted by the mapper. This data
        # is a list of Tuples of (GroupID, Column Name, Value)
        for record in self._map(columns_to_collect):
            # For each aggregation, we need to perform the function against the
            # values as they come in - the collector holds the result up to this
            # point in the set.
            for func, col in aggregations:

                if col != record[1]:
                    continue

                key = f"{func}({col})"

                try:
                    existing = collector.get_value(record[0])
                except KeyError:
                    existing = None
                value = record[2]

                # the aggregation works by performing a simple calculation on
                # the last known value and the value currently seen. This means
                # we don't need a full copy of the data in memory.
                if existing:
                    if value or func == "COUNT":
                        value = AGGREGATORS[func](existing, value)
                    else:
                        value = existing
                elif func == "COUNT":
                    # the COUNT needs seeding with 1, the next cycles are just
                    # adding 1 to the last value.
                    value = 1

                # update the collector with the latest value
                collector.insert(record[0], value)

        # the order of the resulting data set is the order of the hashes - this
        # will appear random, but will ensure the order is consistent between
        # reruns.
        # collector = dict(sorted(collector.items()))

        # We now need to expand out the hashed column names
        for group, results in collector.items():

            for func, col in requested_aggs:
                if func == "AVG":
                    results[f"AVG({col})"] = (
                        results[f"SUM({col})"] / results[f"COUNT({col})"]
                    )

            results = {
                f"{func}({col})": results.get(f"{func}({col})")
                for func, col in requested_aggs
            }

            keys = self._group_keys[group]
            for key in keys:
                results[key[0]] = key[1]

            yield results

    def count(self):
        """
        Count the number of items in each group.

        Yields:
            Dictionary
        """
        # COUNT is a little different, it doesn't have any fields to perform the
        # aggregation on.
        # This implementation could be improved by taking a copy of the
        # aggregate() function and removing the bits that aren't needed to just
        # count the values.
        return self.aggregate(("COUNT", "*"))

    def groups(self):
        """
        Return the set of groups - this is similar to a DISTINCT function
        """
        collector = defaultdict(dict)
        for record in self._map("*"):
            collector[record[0]] = 1
        for group in self._group_keys:
            yield dict(self._group_keys[group])



from .base_plan_node import BasePlanNode

from .aggregate_node import AggregateNode  # aggregate data

# - Evaluate ()
from .group_node import GroupNode  # perform group by
from .distinct_node import DistinctNode  # remove duplicate records
from .dataset_reader_node import DatasetReaderNode  # read datasets

# - JoinNode (currently only INNER JOIN)
from .limit_node import LimitNode  # select the first N records

# - Merge - append sets to each other
from .offset_node import OffsetNode # skip a number of records
from .projection_node import ProjectionNode  # remove unwanted columns including renames
from .selection_node import SelectionNode  # filter unwanted rows

# - SortNode (order a relation by given keys)



import numpy as np
from setuptools import setup
from setuptools import find_packages
from setuptools import Extension
from Cython.Build import cythonize


with open("opteryx/version.py", "r") as v:
    vers = v.read()
exec(vers)  # nosec

with open("README.md", "r") as rm:
    long_description = rm.read()

with open("requirements.txt") as f:
    required = f.read().splitlines()

extensions = [
    Extension(
        name="cjoin",
        sources=["opteryx/third_party/pyarrow_ops/cjoin.pyx"],
        include_dirs=[np.get_include()],
    ),
    Extension(
        name="cythonize",
        sources=["opteryx/third_party/accumulation_tree/accumulation_tree.pyx"],
    )
    #    "mabel/data/internals/group_by.py",
    #    "mabel/data/internals/dictset.py",
    #    "mabel/data/internals/expression.py",
    #    "mabel/data/readers/internals/inline_evaluator.py",
    #    "mabel/data/readers/internals/parallel_reader.py",
    #    "mabel/data/internals/relation.py",
    #    "mabel/data/internals/bloom_filter.py",
    #    "mabel/utils/uintset/uintset.py"
]

setup(
    name="opteryx",
    version=__version__,
    description="Distributed SQL Engine",
    long_description=long_description,
    long_description_content_type="text/markdown",
    maintainer="Joocer",
    author="joocer",
    author_email="justin.joyce@joocer.com",
    packages=find_packages(include=["opteryx", "opteryx.*"]),
    url="https://github.com/mabel-dev/opteryx/",
    install_requires=required,
    ext_modules=cythonize(extensions),
)



# SUPPORT FOR
# ANALYZE TABLE dataset
# [{'Analyze': {'table_name': [{'value': 'be', 'quote_style': None}], 'partitions': None, 'for_columns': False, 'columns': [], 'cache_metadata': False, 'noscan': False, 'compute_statistics': False}}]

# FUTURE SUPPORT FOR
# ANALYZE TABLE dataset PARTITION(date = '2021-01-02');
# [{'Analyze': {'table_name': [{'value': 'be', 'quote_style': None}], 'partitions': [{'BinaryOp': {'left': {'Identifier': {'value': 'date', 'quote_style': None}}, 'op': 'Eq', 'right': {'Value': {'SingleQuotedString': '2021-01-02'}}}}], 'for_columns': False, 'columns': [], 'cache_metadata': False, 'noscan': False, 'compute_statistics': False}}]
